@@ -87,7 +87,7 @@ const char *taktuk_error_msg(int msg_code);
     \see taktuk_error_msg
     \see taktukcomm(3)
 */
-int taktuk_init_threads();
+int taktuk_init_threads(void);
 
 /** Finalize the TakTuk C-interface if threads are used 
     taktuk_leave_threads() attemps to release resources used by the C-interface
@@ -98,7 +98,7 @@ int taktuk_init_threads();
     \see taktuk_init_threads
     \see taktukcomm(3)
 */
-int taktuk_leave_threads();
+int taktuk_leave_threads(void);
 
 /** Gets TakTuk information
     taktuk_get() attemps to get from the TakTuk engine some global information.
