@@ -18,13 +18,13 @@
 #  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA #
 #                                                                             #
 #  Contact: Guillaume.Huard@imag.fr                                           #
-#           ENSIMAG - Laboratoire ID                                          #
+#           ENSIMAG - Laboratoire LIG                                         #
 #           51 avenue Jean Kuntzmann                                          #
 #           38330 Montbonnot Saint Martin                                     #
 #                                                                             #
 ###############################################################################
 
-package taktuk;
+package TakTuk;
 
 # Turns autoflush on for the given filehandle
 sub no_flush($);
@@ -60,13 +60,19 @@ sub error_msg($);
 sub send(%);
 # Receive a message from another node under the form ($to, $from, $message)
 sub recv(%);
-# Decode a string specifying a set of intervals and transforms it to an array
-sub decode_set($);
-# Encode an array of intervals into a compact string specification
-sub encode_set(@);
+# Waits for one of the given TakTuk messages. Returns its code and body
+sub wait_message(@);
+# Get some info from TakTuk
+sub get($);
+ 
+BEGIN
+  {
+    die "FATAL : Perl interpreter too old ($]), Taktuk require Perl >= 5.6.1\n"
+        if ($] < 5.006001);
+  }
 
-our $VERSION = "3.2.2";
-our $RELEASE = sprintf "%d", q$Rev: 288 $ =~ /(\d+)/g;
+our $VERSION = "3.6.3";
+our $RELEASE = sprintf "%d", q$Rev: 477 $ =~ /(\d+)/g;
 # syscalls default granularity
 our $read_size = 32768;
 our $write_size = 32768;
@@ -76,15 +82,26 @@ our $error = undef;
 our $action="A";
 our $eof="D";
 our $taktuk_perl="E";
+our $get="G";
+our $invalid="H";
+our $info="I";
+our $option="N";
 our $timeout="O";
+our $put="P";
+our $reduce_result="Q";
 our $reduce="R";
 our $spread="S";
 our $taktuk_code="T";
+our $update_failed="U";
+our $options="V";
 our $wait_message="W";
+our $resign="X";
 our $arguments="a";
 our $broadcast="b";
-our $downcast="d";
+our $down="d";
 our $execute="e";
+our $file="f";
+our $get_info="g";
 our $input="i";
 our $kill="k";
 our $message="m";
@@ -98,6 +115,7 @@ our $send_to="t";
 our $forward_up="u";
 our $work="w";
 our $synchronize="x";
+our $pipe="z";
 
 # Reduce types
 our $reduce_count = 'c';
@@ -108,7 +126,7 @@ our $reduce_tree = 't';
 ### dispatching of connectors execution     ###
 ###############################################
 
-package taktuk;
+package TakTuk;
 use strict; use bytes;
 
 our %buffer;
@@ -201,12 +219,13 @@ sub syswrite ($$)
                 # system
                 # In any case the message server is blocked and it would be
                 # better to find another solution
-                print STDERR "Delayed write ...";
-                sleep 1;
+                #print STDERR "Delayed write ... $total_expected more to go";
+                #sleep 1;
               }
             else
               {
                 # This is more serious, here we probably eventually end badly
+                print STDERR "Unrecoverable write error\n";
                 $unrecoverable = 1;
                 # I guess we should end up even more badly by killing any child
                 # connector (to avoid partially deployed crashed instances)
@@ -248,7 +267,7 @@ sub get_message ($)
 
     if (exists($buffer{$descriptor}))
       {
-        my ($message, $new_buffer) = taktuk::unpack($buffer{$descriptor});
+        my ($message, $new_buffer) = TakTuk::unpack($buffer{$descriptor});
 
         if (defined($new_buffer))
           {
@@ -324,55 +343,47 @@ our $control_channel_write;
 
 if ($ENV{TAKTUK_CONTROL_READ})
   {
-    open($control_channel_read, "<&=", $ENV{TAKTUK_CONTROL_READ})
+    open($control_channel_read, "<&=". $ENV{TAKTUK_CONTROL_READ})
         or print("Error opening taktuk control channel : $!\n");
+    binmode($control_channel_read);
   }
 if ($ENV{TAKTUK_CONTROL_WRITE})
   {
-    open($control_channel_write, ">&=", $ENV{TAKTUK_CONTROL_WRITE})
+    open($control_channel_write, ">&=". $ENV{TAKTUK_CONTROL_WRITE})
         or print("Error opening taktuk control channel : $!\n");
     no_flush($control_channel_write);
   }
 
-use constant ESWRIT=>0;
-use constant EFCLSD=>1;
-use constant ESREAD=>2;
-use constant EARGTO=>3;
-use constant EARGBD=>4;
-use constant ETMOUT=>5;
-use constant EINVST=>6;
+use constant ESWRIT=>1;
+use constant EFCLSD=>2;
+use constant ESREAD=>3;
+use constant EARGTO=>4;
+use constant EARGBD=>5;
+use constant ETMOUT=>6;
+use constant EINVST=>7;
+use constant EINVAL=>8;
+use constant ENOERR=>9;
+
+our @taktuk_errors = (
+  '"TakTuk::syswrite failed, system message : $!"',
+  '"TakTuk engine closed the communication channel"',
+  '"sysread error, system message : $!"',
+  '"field \"to\" not defined"',
+  '"field \"body\" not defined"',
+  '"timeouted"',
+  '"invalid destination set specification"',
+  '"invalid field required in get"',
+  '"no error"');
+  
 
 sub error_msg($)
   {
     my $error = shift;
 
-    if ($error == ESWRIT)
+    $error--;
+    if ($error <= $#taktuk_errors)
       {
-        return "taktuk::syswrite failed, system message : $!";
-      }
-    elsif ($error == EFCLSD)
-      {
-        return "TakTuk engine closed the communication channel";
-      }
-    elsif ($error == ESREAD)
-      {
-        return "sysread error, system message : $!";
-      }
-    elsif ($error == EARGTO)
-      {
-        return "field 'to' not defined";
-      }
-    elsif ($error == EARGBD)
-      {
-        return "field 'body' not defined";
-      }
-    elsif ($error == ETMOUT)
-      {
-        return "timeouted";
-      }
-    elsif ($error == EINVST)
-      {
-        return "invalid destination set specification";
+        return eval($taktuk_errors[$error]);
       }
     else
       {
@@ -385,31 +396,29 @@ sub send(%)
     my %argument = @_;
     my $from = $ENV{TAKTUK_RANK};
 
-    $error=EARGTO && return undef if not exists($argument{to});
-    my $to = $argument{to};
-    $error=EARGBD && return undef if not exists($argument{body});
-    my $body = $argument{body};
-
-    my @send_set = taktuk::decode_set($to);
-    if (scalar(@send_set))
+    if (not exists($argument{to}))
       {
-        $to = taktuk::encode_set(@send_set);
-        my $full_message = taktuk::encode($taktuk::send_to,
-                                      taktuk::pack($to).
-                                      taktuk::encode($taktuk::message,
-                                      taktuk::pack($to).
-                                      taktuk::pack($from).
-                                      $body));
-        my $result = taktuk::syswrite($control_channel_write,
-                                      taktuk::pack($full_message));
-        $error=ESWRIT if not $result;
-        return $result?$result:undef;
-      }
-    else
-      {
-        $error=EINVST;
+        $error=EARGTO;
         return undef;
       }
+    my $to = $argument{to};
+    if (not exists($argument{body}))
+      {
+        $error=EARGBD;
+        return undef;
+      }
+    my $body = $argument{body};
+
+    my $full_message = TakTuk::encode($TakTuk::send_to,
+                                  TakTuk::pack($to).
+                                  TakTuk::encode($TakTuk::message,
+                                  TakTuk::pack($to).
+                                  TakTuk::pack($from).
+                                  $body));
+    my $result = TakTuk::syswrite($control_channel_write,
+                                  TakTuk::pack($full_message));
+    $error=ESWRIT if not $result;
+    return $result?$result:undef;
   }
 
 sub recv(%)
@@ -423,13 +432,13 @@ sub recv(%)
     # already there (we have to count waiters internally)
     if (exists($argument{timeout}))
       {
-        $message = taktuk::encode($taktuk::wait_message, $argument{timeout});
+        $message = TakTuk::encode($TakTuk::wait_message, $argument{timeout});
       }
     else
       {
-        $message = $taktuk::wait_message;
+        $message = $TakTuk::wait_message;
       }
-    $result = taktuk::syswrite($control_channel_write,taktuk::pack($message));
+    $result = TakTuk::syswrite($control_channel_write,TakTuk::pack($message));
     if (not $result)
       {
         $error=ESWRIT;
@@ -437,115 +446,110 @@ sub recv(%)
       }
 
     # Now we actually get the message
-    $message = get_message($control_channel_read);
-    while ($result and not $message)
-      {
-        $result = read_data($control_channel_read);
-        $message = get_message($control_channel_read);
-      }
+    my $message_code;
+    ($message_code,$message) = wait_message($TakTuk::timeout,$TakTuk::message);
 
-    if ($result)
+    if (defined($message_code))
       {
-        my ($message_code, $to, $from);
-        ($message_code, $message) = taktuk::decode($message);
-        if ($message_code eq $taktuk::timeout)
+        my ($to, $from);
+        if ($message_code eq $TakTuk::timeout)
           {
             $error=ETMOUT;
             return ();
           }
-        ($to, $message) = taktuk::unpack($message);
-        ($from, $message) = taktuk::unpack($message);
+        ($to, $message) = TakTuk::unpack($message);
+        ($from, $message) = TakTuk::unpack($message);
 
         return ($to, $from, $message);
       }
     else
       {
-        if (defined($result))
-          {
-            $error=EFCLSD;
-          }
-        else
-          {
-            $error=ESREAD;
-          }
         return ();
       }
   }
 
-sub decode_set($)
+our @messages;
+
+sub wait_message(@)
   {
-    my $string = shift;
-    my @result = ();
-    my $interval;
-    my $i;
+    my @codes = @_;
+    my ($code, $body);
+    my $result = 1;
+    my $message;
 
-    ($interval,$string) = split /\//,$string,2;
-    while ($interval)
+    for (my $i=0; $i<$#messages; $i+=2)
       {
-        my ($min, $max) = split /-/,$interval,2;
-        my @replacement = ();
-        my $state = 0;
-        my $length = 0;
-
-        $i = $#result;
-        $max = $min if not $max;
-        return () if ($min !~ /^\d+$/o) or ($max !~ /^\d+$/o) or
-                     ($min > $max);
-        while (($i > -1) and ($max < $result[$i]))
+        foreach my $message_code (@codes)
           {
-            $i--;
-            $state = 1 - $state;
-          }
-        unshift(@replacement, $max) if not $state;
-        while (($i > -1) and ($min <= $result[$i]))
-          {
-            $i--;
-            $state = 1 - $state;
-            $length++;
-          }
-        unshift(@replacement, $min) if not $state;
-        $i++;
-        #print "splice @result, $i, $length, @replacement;\n";
-        splice @result, $i, $length, @replacement;
-
-        if (defined($string))
-          {
-            ($interval,$string) = split /\//,$string,2;
-          }
-        else
-          {
-            $interval = 0;
+            if ($messages[$i] eq $message_code)
+              {
+                ($code, $body) = ($messages[$i], $messages[$i+1]);
+                splice @messages, $i, 2;
+                return ($code, $body);
+              }
           }
       }
-    for ($i=1; $i<$#result; $i+=2)
+    while ($result)
       {
-        splice @result, $i, 2
-            while (($i<$#result) and ($result[$i]+1 == $result[$i+1]));
+        $message = get_message($control_channel_read);
+        while ($message)
+          {
+            ($code, $body) = TakTuk::decode($message);
+            foreach my $message_code (@codes)
+              {
+                return ($code, $body) if ($message_code eq $code);
+              }
+            push @messages, $code, $body;
+            $message = get_message($control_channel_read);
+          }
+        $result = read_data($control_channel_read);
       }
-    return @result;
+    if (defined($result))
+      {
+        $error=EFCLSD;
+      }
+    else
+      {
+        $error=ESREAD;
+      }
+    return ();
   }
 
-sub encode_set(@)
+sub get($)
   {
-    my @set = @_;
-    my $result = "";
+    my $result;
+    my $message;
 
-    while (scalar(@set))
+    # Infos query
+    $message = TakTuk::encode($TakTuk::get_info, shift);
+    $result = TakTuk::syswrite($control_channel_write,TakTuk::pack($message));
+    if (not $result)
       {
-        my $min = shift @set;
-        my $max = shift @set;
+        $error=ESWRIT;
+        return -1;
+      }
 
-        $result .= "/" if ($result);
-        if ($min == $max)
+    # Now we actually get the message
+    my $message_code;
+    ($message_code,$message) = wait_message($TakTuk::info,$TakTuk::invalid);
+
+    if (defined($message_code))
+      {
+        if ($message_code eq $TakTuk::invalid)
           {
-            $result .= "$min";
+            $error=EINVAL;
+            return -1;
           }
         else
           {
-            $result .= "$min-$max";
+            $error=ENOERR;
           }
+        return $message;
+      }
+    else
+      {
+        return -1;
       }
-    return $result;
   }
 
 ###############################################
@@ -553,3 +557,189 @@ sub encode_set(@)
 ###############################################
 
 1;
+###############################################################################
+#                                                                             #
+#  TakTuk, a middleware for adaptive large scale parallel remote executions   #
+#  deployment. Perl implementation, copyright(C) 2006 Guillaume Huard.        #
+#                                                                             #
+#  This program is free software; you can redistribute it and/or modify       #
+#  it under the terms of the GNU General Public License as published by       #
+#  the Free Software Foundation; either version 2 of the License, or          #
+#  (at your option) any later version.                                        #
+#                                                                             #
+#  This program is distributed in the hope that it will be useful,            #
+#  but WITHOUT ANY WARRANTY; without even the implied warranty of             #
+#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
+#  GNU General Public License for more details.                               #
+#                                                                             #
+#  You should have received a copy of the GNU General Public License          #
+#  along with this program; if not, write to the Free Software                #
+#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA #
+#                                                                             #
+#  Contact: Guillaume.Huard@imag.fr                                           #
+#           ENSIMAG - Laboratoire LIG                                         #
+#           51 avenue Jean Kuntzmann                                          #
+#           38330 Montbonnot Saint Martin                                     #
+#                                                                             #
+###############################################################################
+
+=pod TakTuk communication layer interface documentation (Perl interface)
+
+=begin html
+
+<center><h1>USER MANUAL</h1></center>
+
+=end html
+
+=head1 NAME
+
+TakTuk - Perl module that provides an interface to C<taktuk(1)> communication
+facilities
+
+=head1 SYNOPSIS
+
+  use TakTuk;
+  
+  my $rank = TakTuk::get('rank');
+  my $count = TakTuk::get('count');
+  
+  print "I'm process $rank among $count\n";
+  
+  if ($rank > 1) {
+      my ($to, $from, $message) = TakTuk::recv();
+      if (not defined($message)) {
+          print "Trying to recv: ",
+                TakTuk::error_msg($TakTuk::error), "\n";
+      } else {
+          print "$to received $message from $from\n";
+      }
+  }
+  
+  sleep 1;
+  my $next = $rank+1;
+  $next = 1 if ($next > $count);
+  if (not TakTuk::send(to=>$next, body=>"[Salut numero $rank]")) {
+      print "Trying to send to $next: ",
+            TakTuk::error_msg($TakTuk::error), "\n";
+  }
+  
+  if ($rank == 1) {
+      my ($to, $from, $message) = TakTuk::recv(timeout=>5);
+      if (not defined($message)) {
+          print "Trying to recv :",
+                TakTuk::error_msg($TakTuk::error), "\n";
+      } else {
+          print "$to received $message from $from\n";
+      }
+  }
+
+=head1 DESCRIPTION
+
+The B<TakTuk> communication layer Perl interface provides a way for programs
+executed using the C<taktuk(1)> command to exchange data. It is based on a
+simple send/receive model using multicast-like sends and optionally timeouted
+receives.  This is only designed to be a control facility, in particular this
+is not a high performance communication library.
+
+WARNING: the B<TakTuk> communication interface is not process-safe : it is
+probably a very bad idea to use point-to-point communication in more than one
+process related to a single B<TakTuk> instance (these processes might be
+several local commands or forked processes).
+
+The Perl communication interface for B<TakTuk> is made of functions that can be
+called by scripts executed using the C<taktuk_perl> command of the B<TakTuk>
+engine (prefered way, less installation requirements on remote machines) or
+using the B<TakTuk> Perl module provided with the B<TakTuk> distribution.
+These functions are:
+
+=over
+
+=item B<TakTuk::get($)>
+
+gets some information from B<TakTuk>. Currently available informations are
+'rank', 'count', 'father', 'child_min' and 'child_max". This is a better way to
+get these informations than environment variables as its takes into account
+renumbering that might occur after process spawn.
+
+=item B<TakTuk::send(%)>
+
+sends a scalar to a single peer or a set specification (see C<taktuk(1)> for
+informations about set specifications).
+The two mandatory fields in the arguments are C<to> (with a set specification)
+and C<body>. Returns an undefined value upon error.
+
+=item B<TakTuk::recv(%)>
+
+blocks until the reception of a message. Returns a list of three elements: the
+logical number of the destination of the message, the logical number of its
+source and the message itself.
+Accepts an optional C<timeout> argument with a numeric value.
+Returns an empty list upon error.
+
+=back
+
+When an error occur, both of these functions set the variable C<$TakTuk::error>
+to the numeric code of the error that occured. A textual description of the
+error is provided by the function C<TakTuk::error_msg($)> that takes the error
+code as an argument.
+
+Error codes are the following :
+
+=over
+
+=item TakTuk::ESWRIT
+
+a call to C<TakTuk::syswrite> failed. This is due to a C<syswrite> error
+different than C<EAGAIN>. The code should be accessible using C<$!>.
+
+=item TakTuk::EFCLSD
+
+the communication channel to the B<TakTuk> engine has been closed. This
+typically occur when shutting down the logical network (using Ctrl-C on root
+node for instance).
+
+=item TakTuk::ESREAD (C<TakTuk::recv> only)
+
+a call to C<sysread> failed (the code should be accessible using C<$!>).
+
+=item TakTuk::EARGTO (C<TakTuk::send> only)
+
+C<to> field missing in the arguments.
+
+=item TakTuk::EARGBD (C<TakTuk::send> only)
+
+C<body> field missing in the arguments.
+
+=item TakTuk::ETMOUT (C<TakTuk::recv> only)
+
+The call to C<TakTuk::recv> timeouted. This only occur when giving a C<timeout>
+field as C<TakTuk::recv> argument.
+
+=begin comment
+
+=item TakTuk::EINVST (C<TakTuk::send> only)
+
+The set specification given as a destination to the C<TakTuk::send> function is
+not correct.
+
+=end comment
+
+=back
+
+=head1 SEE ALSO
+
+C<tatkuk(1)>, C<taktukcomm(3)>, C<TakTuk::Pilot(3)>
+
+=head1 AUTHOR
+
+The original concept of B<TakTuk> has been proposed by Cyrille Martin in his PhD thesis. People involved in this work include Jacques Briat, Olivier Richard, Thierry Gautier and Guillaume Huard.
+
+The author of the version 3 (perl version) and current maintainer of the package is Guillaume Huard.
+
+=head1 COPYRIGHT
+
+The C<TakTuk> communication interface library is provided under the terms
+of the GNU General Public License version 2 or later.
+
+=cut
+
